# changes for .02: make in:* and from:* the default


package Xchat::b0at::thermodynambot;
# 
# No modifications to the code are necessary to use it.
# 
use strict;
use warnings;
use POSIX 'strftime';
#use Data::Dump 'dump';
{ # manually "import" Xchat:: functions for easy use later
  # "use Xchat 'foo';"-style importing only works in 2.4.0+
  # sub/const name is exactly the same unless specified otherwise
	*register       = \&Xchat::register;
	*hook_server    = \&Xchat::hook_server;
	*hook_command   = \&Xchat::hook_command;
	*hook_print     = \&Xchat::hook_print;
	*hook_timer     = \&Xchat::hook_timer;
	*unhook         = \&Xchat::unhook;
	*xprint         = \&Xchat::print;  # name differs
	*xprintf        = \&Xchat::printf; # name differs
	*command        = \&Xchat::command;
	*commandf       = \&Xchat::commandf;
	*find_context   = \&Xchat::find_context;
	*get_context    = \&Xchat::get_context;
	*set_context    = \&Xchat::set_context;
	*get_info       = \&Xchat::get_info;
	*get_prefs      = \&Xchat::get_prefs;
	*emit_print     = \&Xchat::emit_print;
	*nickcmp        = \&Xchat::nickcmp;
	*get_list       = \&Xchat::get_list;
	*user_info      = \&Xchat::user_info;
	*context_info   = \&Xchat::context_info;
	sub EAT_NONE   (){Xchat::EAT_NONE}
	sub EAT_XCHAT  (){Xchat::EAT_XCHAT}
	sub EAT_PLUGIN (){Xchat::EAT_PLUGIN}
	sub EAT_ALL    (){Xchat::EAT_ALL}
	sub REMOVE     (){0} # for timers for all xchat versions
	sub KEEP       (){1}
}
my $NAME    = 'thermodynambot';
my $VERSION = '005.02';
my $PREFIX  = "\02TDB\02\t";
register($NAME, $VERSION, 'The reaction script.');

use constant CONF_FILE    => '_thermodynambot.conf'; # filename looked for in 'xchatdir'
use constant MACRO_CHAR   => '%'; # for conf macros
use constant SEPARATOR    => qr{\s*:\s*}; # for conf properties
use constant LIST_SEP     => qr{\s*,\s*}; # for conf values
use constant NO_MODE      => 'none'; # mode string for no-mode user
use constant REGEX_FLAG   => '&'; # something that can't start a user mask
use constant NETWORK_SEP  => ':'; # network:#channel in +-in/where

my $patterns; # script-wide config entries stored here
my %COMMANDS = (
	'THERMODYNAMBOT' => [ sub{ hookage(); return EAT_ALL; }, 'Usage: THERMODYNAMBOT, reloads '.CONF_FILE ],
	'TDB'            => [ sub{ hookage(); return EAT_ALL; }, 'Usage: TDB'.        ', reloads '.CONF_FILE ],
	'TDBECHO' => [ \&user_test,
		" <nick!user\@addr> '<mode>' <ispriv> <isaction> <network or server> <message text>\n".
		"	   quotes on mode are required (eg '\@', '+', '')\n".
		"	   ispriv and isaction should be 1 or 0\n".
		"	   TDBTEST runs the commands, TDBECHO prints the commands instead\n" ],
);
$COMMANDS{'TDBTEST'} = [ \&user_test, 'Usage: TDBTEST'.$COMMANDS{'TDBECHO'}->[1] ];
$COMMANDS{'TDBECHO'}->[1] = 'Usage: TDBECHO'.$COMMANDS{'TDBECHO'}->[1];
hook_command( $_, $COMMANDS{$_}->[0], { help_text=>$COMMANDS{$_}->[1] } ) for( keys %COMMANDS );
xprint( "\02$NAME v$VERSION\02 by b0at (use /".(join(', /',keys %COMMANDS)).")" ); # announce

my @privmsg_hooks = (); # for hooking/unhooking events
hookage(1); # do config and hook events for the first time

# handle differences in events, then pass to respond, then to do_commands to execute
# stripping color codes from the nick is key
sub hdl_chan             {
	my ($nick, $text, $mode) = @{$_[0]};
	if( my $cmds = respond({nick=>strip_codes($nick), msg=>$text, mode=>$mode, private=>0, action=>0}) )
	{ do_commands( $cmds, get_context() ) }
	return EAT_NONE;
}
sub hdl_actn             {
	my ($nick, $text) = @{$_[0]};
	if( my $cmds = respond({nick=>strip_codes($nick), msg=>$text, mode=>undef, private=>0, action=>1}) )
	{ do_commands( $cmds, get_context() ) }
	return EAT_NONE;
}
sub hdl_priv             {
	my ($nick, $text) = @{$_[0]};
	if( my $cmds = respond({nick=>strip_codes($nick), msg=>$text, mode=>'',    private=>1, action=>0}) )
	{ do_commands( $cmds, get_context() ) }
	return EAT_NONE;
}
sub user_test            { # handle user command triggering for testing
	my ($word, $eol) = @_;
	if( $$word[6] and $$word[2]=~/^(['"])(.*)\1$/ ) {
		my ($nick,$host) = split '!', $$word[1];
		my $cmds = respond( 
		{
			msg=>$$eol[6], 
			#mode=>$2, 
			private=>$$word[3], 
			action=>$$word[4], 
			network=>$$word[5],
		}, 
		{
			nick=>$nick, 
			host=>$host, 
			prefix=>$2
		});
		if( uc $$word[0] eq 'TDBECHO' ) { xprint("$PREFIX $_") foreach( @$cmds ) }
		else { do_commands($_, get_context()) foreach( @$cmds )	}
	}
	else { xprint( $COMMANDS{TDBTEST}->[1] ) }
	return EAT_ALL; # don't pass to server
}

sub respond              {
	my ($event, $user) = @_; # hash refs
	
	# get some context info (needed for user_info_* below)
	$$event{where}     = get_info('channel'); # get tab name
	$$event{server}    = get_info('server')  || 'offline';
	$$event{network} ||= get_info('network') || $$event{server};
	
	# if we don't already have user info...
	$user = user_info( $$event{nick} ) if not ($user and $$user{host});         # check the current tab
	$user = user_info_tabs( $event )   if not $user;                            # next, check other tabs

	if( not defined $user ) {                                                   # next, do a /who <nick>
		user_info_who($event);
		return; # will pick up after who results
	}

	@$user{qw(user addr)} = split '@', $$user{host};
	$$user{who}           = ($$user{nick}||'') . '!' . ($$user{host}||''); # PARENTHESES CRUCIALLY CRITICAL
	my @commands; # store commands to run later

	# check patterns against event data
	for my $rule (@$patterns) {
		
		next if not my @match = $$event{msg} =~ /$$rule{pattern}/;              # required text

		# action/private + network
		for my $type ('action', 'private') { if($$event{$type}) {
			next if     match($$event{network}, $$rule{$type}[0]);               # excluded net:type combos
			next unless match($$event{network}, $$rule{$type}[1]);               # required net:type combos
		}}

		# people
		next if     match($$user{who}, $$rule{from}[0]);                        # excluded people
		next unless match($$user{who}, $$rule{from}[1]);                        # required people

		# public events
		if( not $$event{private} )
	{ 
		# channels + network
		next if match($$event{where}, $$rule{where}[0]);                        # excluded channels
		next if match_mult([@$event{qw(network where)}], $$rule{network}[0]);   # excluded net:chan combos
		
		next unless match($$event{where}, $$rule{where}[1])                     # allowed channels
			or match_mult([@$event{qw(network where)}], $$rule{network}[1]);    # allowed net:chan combos

		# modes
		if( not $$user{prefix} and not $$rule{nomode}[1] ) {          # if user modeless + modeless not explicitly allowed
			next if $$rule{nomode}[0];                                 # ... next if modeless explicitly disallowed
			next if not $$rule{nomode}[1] and not $$rule{nomode}[1]    # ... next if modeless not explicitly mentioned
				and scalar @{$$rule{mode}[1]}                           # ... + and other modes are required
		
		} elsif( $$user{prefix} ) {                                   # elsif user has a mode
			next if not @{$$rule{mode}[1]} 	                          # ... next if no modes are required
				and $$rule{nomode}[1];                                  # ... + except for modeless
			next if scalar @{$$rule{mode}[1]}                          # ... next if any modes are required
				and not match_eq($$user{prefix}, $$rule{mode}[1]);      # ... + and user doesn't match one
			next if scalar @{$$rule{mode}[0]}                          # ... next if any modes are excluded
				and match_eq($$user{prefix}, $$rule{mode}[0]);          # ... + and user matches one
		}
	}
		# the event matched
		push @commands, expand_macros( $$rule{commands}, \@match, $event, $user);
	}
	return \@commands;
}
sub load_conf        ($) { # read in conf, return valid data; print on error, stop on fatal error
	my ($filename, $count, $fh, $new_conf) = (shift, 0,);
	return unless( $filename and -e $filename 
		and open $fh, $filename );

	my @p; # store pattern entries here
	while( my $l = <$fh> ) {
		++$count; # line count
		
		next if $l=~/^#/;
		$l =~ s/^\s+//; # trim after checking for comments, '#' must be first char
		$l =~ s/\s+$//;
		next if $l=~/^$/;
		last if $l=~/^__END__$/;
		
		if( my ($method, $pattern)     = $l =~ /^\[(text|regex|line)(??{SEPARATOR})(.+?)\]$/i ) {
			# new entry
			push @p, { #   => [ neg (0),  pos (1) ],
					#from     => [ [],       []      ], # people
					#where    => [ [],       []      ], # places
					
					# 005.02: DEFAULT to +in: *
					from     => [ [],       ['.+']  ], 
					where    => [ [],       ['.+']  ],
					
					network  => [ [],       []      ], # places w/network specified
					mode     => [ [],       []      ], # status
					# special values, bool
					nomode   => [ undef,    undef   ], # there's probably a better way, but it's fine for now
					#			networks allowed, networks disallowed
					private  => [ [],       []      ],
					action   => [ [],       []      ],
					commands => [],							# just a list
				};
			
			if( lc $method eq 'text' ) {   
				$p[$#p]{pattern} = '(?i)'.wildcard($pattern); # '(?i)' makes it case insensitive
			}
			elsif( lc $method eq 'line' )  {
				$p[$#p]{pattern} = '(?i)^'.wildcard($pattern).'$'; # new to 005
			}
			elsif( lc $method eq 'regex' ) {
				$p[$#p]{pattern} = $pattern;
			}
		}
		elsif ( $l =~ /^\[.*\]$/i ) { # bad form in an entry, don't risk overlapping entries' data
			xprint("$PREFIX Parse error: '$_' is not of the form '[method:REGULAR EXPRESSION]',");
			xprint("$PREFIX \cBParse halted.\cB");
			last;
		}

		elsif( my ($where, $wherelist) = $l =~ /^(\+|-)(?:in|where)(??{SEPARATOR})(.+)$/i )     {
			$where = ($where eq '+') ? 1 : 0;
			for my $place ( split LIST_SEP, $wherelist ) {
				my ($net, $type, $chan);
				
				if( ($net, $type) = $place =~ 
					/^
						(?: 
							(.+?)               # network name
							(??{NETWORK_SEP})
						)?                     # ... is optional
						(action|private)       # non-channel type of message ($type)
					$/ix )
				{
					if( not $net ) {  # 'private' or 'action' without a network
						push @{ $p[$#p]{ lc $type }[$where] }, '(?i)'.wildcard('*');  # set allowed/disallowed for all networks
					}
					else { # network:private or network:action
						push @{ $p[$#p]{ lc $type }[$where] }, '(?i)'.wildcard($net);
					}
				}			
				# network specified, de-wildcard and stick in an array ref
				elsif( ($net, $chan) = $place =~ 
					/^
						(.+?)                  # network name
						(??{NETWORK_SEP})
						(.+?)                  # channel name
					$/x )
				{	# a different structure for the 'network' key: [net-wildcard, chan-wildcard]
					push @{ $p[$#p]{network}[$where] }, [ '(?i)'.wildcard($net), '(?i)'.wildcard($chan) ];
				}
				else { # regular channel wildcard
					push @{ $p[$#p]{where  }[$where] }, '(?i)'.wildcard($place);
				}
			}
		}
		elsif( my ($from,  $fromlist)  = $l =~ /^(\+|-)from(??{SEPARATOR})(.+)$/i )             {
			$from = ($from eq '+') ? 1 : 0;
			for ( split LIST_SEP, $fromlist ) {
				if( /^(??{REGEX_FLAG})([^!]+![^@]+@.+)$/ ) { # full mask - regex
					push @{ $p[$#p]{from}[$from] }, $2;
				}
				elsif( /^[^!]+![^@]+@.+$/ )                { # full mask - wildcard
					push @{ $p[$#p]{from}[$from] }, wildcard($_);
				}
				elsif( /^[^!]+!$/ )                        { # lonesome nick ('bob!')
					push @{ $p[$#p]{from}[$from] }, wildcard($_).".*?@.*?";
				}				
				elsif( /^\@.+$/ )                          { # lonesome address ('@foo.com')
					push @{ $p[$#p]{from}[$from] }, ".*?!.*?".wildcard($_);
				}
				else { # unknown - assume wildcard on full mask (eg '*')
					push @{ $p[$#p]{from}[$from] }, wildcard($_);
				}
			}
		}
		elsif( my ($mode,  $modelist)  = $l =~ /^\s*(\+|-)mode(??{SEPARATOR})(.+)$/i )          {
			$mode = ($mode eq '+') ? 1 : 0;
			my %modelist = map {$_=>1} split LIST_SEP, $modelist;
			# extract NO_MODE keyword
			if( $modelist{+NO_MODE} ) { # NO_MODE is a constant
				$p[$#p]{nomode}[$mode] = 1; # eg -mode: none -> nomode[-==0]=1 -> no moders are disallowed
				delete $modelist{+NO_MODE};
			}
			push @{ $p[$#p]{mode}[$mode] }, keys %modelist; # regular modes
		}
		elsif( my ($command)           = $l =~ /^command(??{SEPARATOR})(.+)$/i )                {
			push @{ $p[$#p]{commands} }, $command;
		}
		else { # parse error notice
			xprint("$PREFIX Parse error on line $count: unknown data.");
		}
	} # <$fh>
	close $fh;
	return \@p;
}

# helpers

sub do_commands     ($$) {
	my ($cmds, $context) = @_;
	hook_timer(0, # immediately after this event
	sub {
		#my $original_context = get_context(); # uncomment for polite context switching
		set_context( $context );
		command( $cmds );
		#set_context($original_context); # uncomment for polite context switching
		return REMOVE; # do this timer only once
	} );
}
sub user_info_tabs   ($) {
	my ($event) = @_;
	my $original_context = get_context();
	for my $tab ( get_list('channels') ) {
	
		# find channel on same server as event
		next unless $$tab{type} == 2; # channel
		next unless((
			$$event{network} and $$tab{network} and
				$$tab{network} eq $$event{network}
		) or (
			$$event{server}  and $$tab{server}  and
				$$tab{server}  eq $$event{server}
		));
		
		# channel on same server: go there
		set_context( $$tab{context} );
		
		# find user with same nick
		for my $u ( get_list('users') ) {
			next unless nickcmp( $$u{nick}, $$event{nick} )==0;
			# found the guy
			set_context( $original_context ); # be polite
			# return a shallow copy with a different prefix
			# (if in another channel, prefix doesn't apply)
			return {%$u, prefix=>''};
			# doing $u={%$u}; then setting prefix and returning $u
			#	resulted in some strange "attempt to free unreferenced
			#	scalar" errors, this method works without resorting
			#	to Storable's dclone()
		}
	}
	return;
}
sub user_info_who    ($) {
	my ($event) = @_;
	my ($context, $who, $who_end) = (get_context(), );
	
	# hook /who output
	$who = hook_server('352', sub {
		my ($w, $e) = @_;
		return EAT_NONE if nickcmp($$w[7],$$event{nick})!=0;
		unhook($who); undef($who); # remove this hook
		# nick matches! get commands
		do_commands( respond($event, { # fill in user data...
			away => ($$w[8]=~/G/?1:0),
			nick => $$w[7], # actual, case-correct nick
			host => $$w[4].'@'.$$w[5],
			prefix => $$event{mode}||'', # doesn't apply if they're not in the channel
		}), $context ); # do commands in required context
		return EAT_XCHAT; # hide our /who from user
	} );
	
	# hide "End of /who output" line from user, too
	$who_end = hook_server('315', sub {
		return EAT_NONE if $who; # still waiting on our who output
		unhook($who_end); return EAT_XCHAT;
	} );
	
	# now send the command and wait for server response
	command("who $$event{nick}");
}
sub match           ($$) { # find first /^match$/ in list, return pattern matched for success or undef
	my ($value, $list) = @_;
	for ( @$list ) { return $_ if $value =~ /^$_$/ }
	return;
}
sub match_eq        ($$) { # the '+' for voice mode screws with regex
	my ($value, $list) = @_;
	my %search = map {$_=>1} @$list;
	return exists $search{$value};
}
sub match_mult     ($$$) {
	my ($values, $list) = @_;
	for my $sublist ( @$list ) { # gripes big time if sizes don't match!
		return $sublist if map { $$values[$_] =~ /^$$sublist[$_]$/ } (1..$#$values);
	}
	return;
}
sub wildcard         ($) {
	# quote meta characters coming in
	my $w = "\Q$_[0]\E"; # makes <*> -> <\*>

	# need to find and replace quotmeta'd '\\'.'\*' now, not just '\*'
	$w =~ s!\\\?     !\E(.)\Q!gx;   # one character
	$w =~ s!\\\*(?=.)!\E(.*?)\Q!gx; # non-greedy required here
	$w =~ s!\\\*    $!\E(.*)\Q!gx;  # greedy required here
	# bug fix: using .+ to find chars after '*' was skipping later wildcards,
	# should've used a look-ahead to begin with

	return $w;
}
sub strip_codes      ($) { # strip color codes for print events which laden nicks with them
	my $s = shift;
	$s=~s/\cB//g;
	$s=~s/\cC\d{0,2}(?:,\d{0,2})?//g;
	$s=~s/\cG//g;
	$s=~s/\cO//g;
	$s=~s/\cV//g;
	$s=~s/\c_//g;
	return $s;
}
sub expand_macros ($$$$) {
	# m includes matches and '$' for last match
	my ($cmds, $match, $event, $user) = @_;
	my $split = '('.MACRO_CHAR. '.)';
	my $bit   =     MACRO_CHAR.'(.)';
	return map { join '', map {
		if(/$bit/) { # is a macro
			# match against only non-macro char,
			# don't allow later matches to clobber actual value
			$_ = $1;
			if($_ eq MACRO_CHAR) { +MACRO_CHAR }
			# user data
			elsif(/n/) {$$user{nick}     }
			elsif(/u/) {$$user{user}     }
			elsif(/a/) {$$user{addr}     }
			elsif(/o/) {$$user{prefix}   }
			# event data
			elsif(/s/) {$$event{msg}     }
			elsif(/w/) {$$event{network} }
			elsif(/c/) {$$event{where}   }
			# captures: zero-indexed, macros start at %1, so sub. 1
			elsif(/\d/) {exists $$match[$_-1]     ? $$match[$_-1]     : ''}
			elsif(/\$/) {exists $$match[$#$match] ? $$match[$#$match] : ''}
			# color codes
			elsif(/B/) {"\cB"} # %B == \cB
			elsif(/C/) {"\cC"} # %C == \cC
			elsif(/O/) {"\cO"} # %O == \cO
			elsif(/R/) {"\cV"} # %R == \cV
			elsif(/U/) {"\c_"} # %U == \c_
			# own nick
			elsif(/m/) {get_info('nick')}
			# time/date
			elsif(/d/) {strftime('%Y/%m/%d', localtime)} # local date
			elsif(/y/) {strftime('%Y/%m/%d', gmtime   )} # zulu date
			elsif(/t/) {strftime('%H:%M:%S', localtime)} # local time
			elsif(/z/) {strftime('%H:%M:%S', gmtime   )} # zulu time
			else {''}
		} else {$_} # is not a macro
	} split /$split/ } @$cmds;
}
sub hookage              { # get full conf path, attempt to load it, hook or unhook event
	my $quiet = shift;
	my $conffile = get_info('xchatdir');
	$conffile =~ s!\\+!/!g; # de-windows-ize it if necessary
	$conffile =~ s!/$!!; # remove ending slash to prevent doubling up, just in case
	$conffile .= '/'.CONF_FILE;

	if( $patterns = load_conf($conffile) )
	{
		if( not @privmsg_hooks )
		{
			push @privmsg_hooks, hook_print('Channel Message',		   \&hdl_chan);
			push @privmsg_hooks, hook_print('Channel Msg Hilight',	   \&hdl_chan);
			push @privmsg_hooks, hook_print('Channel Action',			\&hdl_actn);
			push @privmsg_hooks, hook_print('Channel Action Hilight',	\&hdl_actn);
			push @privmsg_hooks, hook_print('Private Message',		   \&hdl_priv);
			push @privmsg_hooks, hook_print('Private Message to Dialog', \&hdl_priv);
		}
		xprint("$PREFIX Config loaded, events hooked.") if not $quiet;
	}
	else
	{
		if( @privmsg_hooks )
		{
			unhook($_) for @privmsg_hooks; # unhook all
			@privmsg_hooks = (); # set it to empty
		}
		xprint("$PREFIX \02Error\02: Config unable to load. Events not hooked.");
	}
}

1;
__END__
